#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/xpm.h>
#include <sys/time.h>
#include <signal.h>
#include <values.h>

const char *ProgramName="xmuwin";
const char *Version="version 1.01";
const char *Copyright="graphics by Mominoha (C) 1998,1999.\nMS-Windows version programmed by Narui (C) 1998,1999.\nX11 version programmed by Nakatsuka (C) 1999.";

#include "mu0.xpm"
#include "mu1.xpm"
#include "mu2.xpm"
#include "mu3.xpm"
#include "mu4.xpm"
#include "mu5.xpm"
#include "mu6.xpm"
#include "mu7.xpm"

#define PIXMAPNUM 8
#define PIXMAPSIZEX 105
#define PIXMAPSIZEY 193
#define PIXMAPCENTERX 75
#define PIXMAPCENTERY 60
#define MINWINSIZEX 30
#define MINWINSIZEY 133

#ifdef ENABLEALLOCA
#define MALLOC alloca
#define FREE(u)
#else
#define MALLOC malloc
#define FREE(u) free(u)
#endif


typedef void (*sighandler_t)(int);

const int maxptn=85;
const int pixmapptn[]={
  7,6,5,
  0,0,0,0,1,1,1,1,
  0,0,0,0,0,0,0,0,
  0,0,0,0,1,1,1,1,0,0,0,0,1,1,1,1,
  0,0,0,1,1,1,0,0,0,1,1,1,
  0,0,0,0,0,0,0,0,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,
  3,4,4,4,3,
  2,2,2,2,2,2,2,2,
  5,6,7,
};

/* interval is animation speed */
/* NOT probability of appearance */
const int interval[]={
  20000,40000,60000,80000,100000
};
int arg_interval=2;

const int itprobrate[]={
  3000,1500,1000,750,600
};

const int probability[]={
  2,5,10,20,40
};
int arg_probability=2;
int arg_targetoverride=0;

Pixmap pixmap[9],mask[9];
Window window;
char *displayname=NULL;
Display *display;
int screennumber;
GC gc;
int inanim=0;
int animptn=0;
int tcount=0;
Window attachwin;

int loadpixmap(void)
{
  /* create pixmap */
  if(XpmCreatePixmapFromData(display,window,muxpm0,&pixmap[0],&mask[0],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm1,&pixmap[1],&mask[1],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm2,&pixmap[2],&mask[2],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm3,&pixmap[3],&mask[3],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm4,&pixmap[4],&mask[4],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm5,&pixmap[5],&mask[5],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm6,&pixmap[6],&mask[6],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(display,window,muxpm7,&pixmap[7],&mask[7],0)!=XpmSuccess)
    return 0;
  return 1;
}

void releasepixmap(void)
{
  int i;
  for(i=0;i<PIXMAPNUM;i++) {
    XFreePixmap(display,pixmap[i]);
    XFreePixmap(display,mask[i]);
  }
}

int SortWindow(void) {
  Window root,parent,*children;
  int nchildren;
  int i,j;
  XWindowAttributes childattr;
  Window rstack[3];

  /* get window list */
  XQueryTree(display,RootWindow(display,0),
	     &root,
	     &parent,
	     &children,
	     &nchildren);
  
  /* search attached window */
  for(i=0;i<nchildren;i++)
    if(children[i] == attachwin) break;
  if(i==nchildren) {
    /* error */
    XFree(children);
    return 0;
  }
  if(i==nchildren-1) {
    /* attached window is top */
    XRaiseWindow(display,window);
    XFree(children);
    return 1;
  }
  if(children[i+1]==window) {
    XFree(children);
    return 1; /* all right */
  }
  for(j=i+1;j<nchildren;j++) {
    XGetWindowAttributes(display,children[j],&childattr);
    if(childattr.map_state == 2) break;
  }
  if(j==nchildren) {
    /* attached window is top */
    XRaiseWindow(display,window);
    XFree(children);
    return 1;
  }
  if(children[j]==window) {
    XFree(children);
    return 1; /* all right */
  }
  rstack[0]=children[j];
  rstack[1]=window;
  rstack[2]=attachwin;
  XRestackWindows(display,rstack,3);
  XFree(children);
  return 1;
}

int isvalidwindow(Window w)
{
  XWindowAttributes attr;
  XGetWindowAttributes(display,w,&attr);

  if(attr.map_state == 2   /* visible */
     && (arg_targetoverride || !attr.override_redirect)
     && attr.width  > MINWINSIZEX
     && attr.height > MINWINSIZEY /* window size */
     && attr.x > PIXMAPCENTERX
     && attr.y > PIXMAPCENTERY
     && attr.x < DisplayWidth(display,screennumber)-MINWINSIZEX
     && attr.y < DisplayHeight(display,screennumber)-MINWINSIZEY
     /* window pos */
     ) return 1;
  return 0;
}

void animation(int ptnno)
{
  XWindowAttributes attr;

  SortWindow();

  /* window pos check */
  if(!isvalidwindow(attachwin)) {
    inanim=0;
    XUnmapWindow(display,window);
    XFlush( display );
    return;
  }
  
  XGetWindowAttributes(display,attachwin,&attr);
  XMoveWindow(display,window,
	      attr.x-PIXMAPCENTERX,
	      attr.y-PIXMAPCENTERY);

  /* draw pixmap */
  XShapeCombineMask(display,window,
		    ShapeBounding,0,0,mask[pixmapptn[ptnno]],ShapeSet);
  XCopyArea(display,
	    pixmap[pixmapptn[ptnno]],window,gc,0,0,PIXMAPSIZEX,PIXMAPSIZEY,0,0);
  XFlush( display );
}

int GetAttachWindow(Window *w)
{
  Window root,parent,*children;
  Window *cand;
  int nchildren;
  int i;
  int winnum=0;

  /* get window list */
  XQueryTree(display,RootWindow(display,0),
	     &root,
	     &parent,
	     &children,
	     &nchildren);

  /* get valid window */
  cand=(Window *)MALLOC(sizeof(Window)*nchildren);
  if(!cand) {
    fprintf(stderr,"cannot allocate memory.\n");
    signal(SIGALRM,SIG_IGN);
    
    XSetErrorHandler(NULL);
    
    XFreeGC(display,gc);
    
    XUnmapWindow(display,window);
    releasepixmap();
    XDestroyWindow(display,window);
    XCloseDisplay(display);
    exit(0);
  }

  for(i=0;i<nchildren;i++) {
    if(isvalidwindow(children[i])) {
      cand[winnum]=children[i];
      winnum++;
    }
  }

  /* select window */
  if(winnum==0) {
    XFree(children);
    FREE(cand);
    return 0;
  }
  i=(int) ((double)winnum*rand()/(RAND_MAX+1.0));
  *w=cand[i];
  XFree(children);
  FREE(cand);
  return 1;
}

void AlarmHandler(void)
{
  signal(SIGALRM,SIG_IGN);
  if(inanim) {
    animptn++;
    if(animptn == maxptn) {
      inanim=0;
      XUnmapWindow(display,window);
      XFlush(display);
    } else {
      animation(animptn);
    }
  } else {
    if((int)((double) itprobrate[arg_interval]*
	     (double) probability[arg_probability] *
	     rand()/(RAND_MAX+1.0))<tcount &&
       GetAttachWindow(&attachwin)){
      tcount=0;
      XMapWindow( display, window );
      /*XFlush( display );*/
      animptn=0;
      inanim=1;
      animation(animptn);
    } else tcount++;
  }
  signal(SIGALRM,(sighandler_t) AlarmHandler);
}

static int (*NextErrorHandler)(Display *d, XErrorEvent *ev);

static int ErrorHandler(Display *d, XErrorEvent *ev)
{
  if (ev->error_code == BadWindow) {
    inanim=0;
    XUnmapWindow(display,window);
    return 0;
  }
  else    
    return NextErrorHandler(d,ev);
}

int ParseArgument(int argc,char *argv[])
{
  int i;
  for(i=1;i<argc;i++) {
    if (!strcmp(argv[i],"-display")) {
      if( i + 1 >= argc ) {
	fprintf(stderr,"display is not specified.\n");
	return 0;
      }
      displayname=argv[i+1];
      i++;
    } else if (argv[i][0] == '-' && argv[i][1] == 's') {
      if (strlen(argv[i]) == 3 && argv[i][2] >= '0' && argv[i][2] <= '9')
	arg_interval=atoi(argv[i]+2);
      else {
	if( i + 1 >= argc ) {
	  fprintf(stderr,"speed is not specified.\n");
	  return 0;
	}
	arg_interval=atoi(argv[i+1]);
	i++;
      }
      if(arg_interval < 0 || arg_interval > 4) {
	fprintf(stderr,"speed is invalid.\n");
	return 0;
      }
    } else if (argv[i][0] == '-' && argv[i][1] == 'f') {
      if (strlen(argv[i]) == 3 && argv[i][2] >= '0' && argv[i][2] <= '9')
	arg_probability=atoi(argv[i]+2);
      else {
	if( i + 1 >= argc ) {
	  fprintf(stderr,"frequency is not specified.\n");
	  return 0;
	}
	arg_probability=atoi(argv[i+1]);
	i++;
      }
      if(arg_probability < 0 || arg_probability > 4) {
	fprintf(stderr,"frequenct is invalid.\n");
	return 0;
      }
    } else if (argv[i][0] == '-' && argv[i][1] == 'o') {
      arg_targetoverride=1;
    } else if (!strcmp(argv[i],"-help") || !strcmp(argv[i],"-h")) {
      printf("\n%s %s\n",ProgramName,Version);
      printf("%s\n\n",Copyright);
      printf("-help , -h         print this text\n");
      printf("-display <host:n>  specify display\n");
      printf("-s <n>             specify speed.\n");
      printf("-f <n>             specify appearance frequency.\n");
      printf("-o                 enable targetiing override_redirect window.\n");
      printf("\n");
      return 0;
    } else {
      fprintf(stderr,"invalid option : %s\n",argv[i]);
      return 0;
    }
#ifdef DEBUG
    printf("interval:%d\n",arg_interval);
    printf("probability:%d\n",arg_probability);
    printf("targetoverride:%d\n",arg_targetoverride);
#endif
  }
  return 1;
}

int main(int argc,char *argv[])
{
  XSetWindowAttributes attr;
  XEvent event;
  int loop=1;
  struct itimerval itime;

  srand((unsigned)getpid());

  if(!ParseArgument(argc,argv)) return 0;

  if(!(display = XOpenDisplay(displayname))) {
    fprintf(stderr,"cannot open display.\n");
    return 0;
  }
  screennumber=DefaultScreen(display);
  attr.save_under=True;
  attr.override_redirect=True;
  window=XCreateWindow(display,RootWindow(display,screennumber),
		  0,0,
		  PIXMAPSIZEX,PIXMAPSIZEY,
		  0,DefaultDepth(display,screennumber),
		  InputOutput,
		  CopyFromParent,
		  CWSaveUnder | CWOverrideRedirect,
		  &attr);
		  
  if(!loadpixmap()) {
    fprintf(stderr,"can't load pixmaps.\n");
    XDestroyWindow(display,window);
    XCloseDisplay(display);
    return 0;
  }

  gc=XCreateGC(display,window,0,0);
  XSetGraphicsExposures(display,gc,False);

  NextErrorHandler=XSetErrorHandler(ErrorHandler);

  /* event */
  XSelectInput(display,window,ButtonPressMask);

  /* timer handler */
  signal(SIGALRM,(sighandler_t) AlarmHandler);

  timerclear(&itime.it_interval);
  timerclear(&itime.it_value);
  itime.it_interval.tv_usec=interval[arg_interval];
  itime.it_value.tv_usec=interval[arg_interval];
  setitimer(ITIMER_REAL,&itime,0);

  while(loop) {

    /* animation process */
    
    /* event process */
    XNextEvent(display,&event);
    
    switch(event.type) {
    case ButtonPress:
      switch(event.xbutton.button) {
      case 1:
	/* window unmap not to be obstacle */
	inanim=0;
	XUnmapWindow(display,window);
	break;
      case 3:
	/* exit program */
	loop=0;
	break;
      default:
	break;
      }
      break;
    case Expose:
      if(inanim) animation(animptn);
      break;
    default:
      break;
    }
  }
  signal(SIGALRM,SIG_DFL);

  XSetErrorHandler(NULL);

  XFreeGC(display,gc);

  XUnmapWindow(display,window);
  releasepixmap();
  XDestroyWindow(display,window);
  XCloseDisplay(display);
  return 0;
}



